from dataclasses import dataclass
from typing import Optional


@dataclass
class NewWifiConfig:
    ssid_24ghz: Optional[str] = None  # SSID for the 2.4GHz network (Optional)
    password_24ghz: Optional[str] = None  # Password for the 2.4GHz network (Optional)
    ssid_5ghz: Optional[str] = None  # SSID for the 5GHz network (Optional)
    password_5ghz: Optional[str] = None  # Password for the 5GHz network (Optional)
    hide_24ghz: bool = False  # Flag to hide the 2.4GHz network SSID (Optional)
    hide_5ghz: bool = False  # Flag to hide the 5GHz network SSID (Optional)


@dataclass
class CurrentWifiConfig:
    bssid_24ghz: Optional[str] = None  # BSSID for the 2.4GHz network (internal use)
    bssid_5ghz: Optional[str] = None  # BSSID for the 5GHz network (internal use)
    current_ssid_24ghz: Optional[str] = None  # Current SSID for the 2.4GHz network (internal use)
    current_ssid_5ghz: Optional[str] = None  # Current SSID for the 5GHz network (internal use)




import grpc
import time
from spacex.api.device import device_pb2_grpc, device_pb2
from contextlib import contextmanager

from starlink_client.location import Location

STARLINK_LAN_GRPC_SERVER_ADDRESS = "192.168.100.1:9200"
ERR_TIMEOUT = "request timed out"
DEFAULT_TIMEOUT = 5


class GrpcClient:

    def __init__(self, host: str):
        self._host = host
        self._channel = grpc.insecure_channel(host)
        self._client = device_pb2_grpc.DeviceStub(self._channel)

    def call(self, req: device_pb2.Request) -> device_pb2.Response:
        with self._timeout():
            return self._call(req)

    def _call(self, req: device_pb2.Request) -> device_pb2.Response:
        return self._client.Handle(req)

    @contextmanager
    def _timeout(self, seconds: int = DEFAULT_TIMEOUT):
        start = time.time()
        yield
        if time.time() - start > seconds:
            raise TimeoutError(ERR_TIMEOUT)

    def get_location(self) -> Location:
        grpc_req = device_pb2.Request(
            get_location=device_pb2.GetLocationRequest()
        )
        resp: device_pb2.Response = self.call(grpc_req)
        lat = resp.get_location.lla.lat
        lon = resp.get_location.lla.lon
        return Location(
            lat=lat,
            lon=lon,
            alt=resp.get_location.lla.alt
        )



from typing import Optional
import httpx
import time
from contextlib import contextmanager
from spacex.api.device import device_pb2
from http.cookies import SimpleCookie
import threading
import json
import hashlib
import os
from http.cookiejar import Cookie
from starlink_client.account import Account

# API URLs
STARLINK_GRPC_WEB_API_URL = "https://api2.starlink.com/SpaceX.API.Device.Device/Handle"
STARLINK_AUTH_URL = "https://api.starlink.com/auth-rp/auth/user"

# Constants
ERR_TIMEOUT = "request timed out"
DEFAULT_TIMEOUT = 5  # Timeout for requests in seconds
GRPC_PREFIX_SIZE = 5  # gRPC prefix size for framing messages


class AuthenticationError(Exception):
    """Raised when authentication fails."""
    pass


class ResponseError(Exception):
    """Raised for errors in the response from the server."""
    pass


class GrpcWebBaseClient:
    def __init__(self, initial_cookies: str, cookie_storage_path: str):
        """
        Initialize the gRPC-Web client.

        Parameters:
            initial_cookies (str): The initial cookies for authentication.
            cookie_storage_path (str): The path to store cookies on disk.
        """
        self._url = STARLINK_GRPC_WEB_API_URL
        self._auth_url = STARLINK_AUTH_URL
        self._lock = threading.Lock()
        self._xsrf_token: Optional[str] = None

        # Generate a hash for the initial cookie
        self._cookie_hash = hashlib.sha256(initial_cookies.encode('utf-8')).hexdigest()
        os.makedirs(cookie_storage_path, exist_ok=True)
        self._cookie_file = os.path.join(cookie_storage_path, f"{self._cookie_hash}.json")

        # Shared cookie jar for session cookies
        self._cookie_jar = httpx.Cookies()

        # HTTP client initialized con HTTP/2
        self._client = httpx.Client(http2=True, cookies=self._cookie_jar)
        self._cookie = None
        # Intentar cargar las cookies desde el archivo o establecer las cookies iniciales
        if os.path.exists(self._cookie_file):
            self._load_cookies_from_file()
        else:
            self._load_cookies(initial_cookies)
        self._refresh_auth()
        self._save_cookies_to_file()
        # Account object to hold account-related data
        self._account: Optional[Account] = None
        

    def _load_cookies(self, cookies: str):
        """
        Parse and load cookies into the shared cookie jar.

        Parameters:
            cookies (str): Raw cookie string.
        """
        cookie = SimpleCookie()
        cookie.load(cookies)
        self._cookies_dict = {key: morsel.value for key, morsel in cookie.items()}
        self._xsrf_token = self._cookies_dict.get("XSRF-TOKEN", "")
        self._cookie = cookies

        # Update the shared cookie jar with parsed cookies
        for key, value in self._cookies_dict.items():
            self._cookie_jar.set(key, value)

    def _save_cookies_to_file(self):
        """Save cookies to a JSON file to reuse them across sessions."""
        cookies_list = []
        for cookie in self._cookie_jar.jar:
            if isinstance(cookie, Cookie):
                cookies_list.append({
                    'name': cookie.name,
                    'value': cookie.value,
                    'domain': cookie.domain,
                    'path': cookie.path,
                    'expires': cookie.expires,
                    'secure': cookie.secure,
                    'httponly': cookie.has_nonstandard_attr('HttpOnly'),
                    'samesite': cookie.get_nonstandard_attr('samesite')
                })
        with open(self._cookie_file, 'w', encoding='utf-8') as f:
            json.dump(cookies_list, f, ensure_ascii=False, indent=4)

    def _load_cookies_from_file(self):
        """Load cookies from a JSON file if available."""
        with open(self._cookie_file, 'r', encoding='utf-8') as f:
            cookies_list = json.load(f)
        for cookie_dict in cookies_list:
            cookie = Cookie(
                version=0,
                name=cookie_dict.get('name', ''),
                value=cookie_dict.get('value', ''),
                port=None,
                port_specified=False,
                domain=cookie_dict.get('domain', ''),
                domain_specified=bool(cookie_dict.get('domain', '')),
                domain_initial_dot=cookie_dict.get('domain', '').startswith('.'),
                path=cookie_dict.get('path', '/'),
                path_specified=bool(cookie_dict.get('path', '/')),
                secure=cookie_dict.get('secure', False),
                expires=cookie_dict.get('expires', None),
                discard=False,
                comment=None,
                comment_url=None,
                rest={'HttpOnly': cookie_dict.get('httponly', False)},
                rfc2109=False,
            )
            self._cookie_jar.jar.set_cookie(cookie)
        self._xsrf_token = self._cookie_jar.get("XSRF-TOKEN", "")
        self._cookie = "; ".join([f"{cookie.name}={cookie.value}" for cookie in self._cookie_jar.jar])

    def call(self, req: device_pb2.Request) -> device_pb2.Response:
        """
        Make a gRPC-Web call and handle authentication errors.

        Parameters:
            req (device_pb2.Request): The request to send.

        Returns:
            device_pb2.Response: The server response.
        """
        with self._timeout():
            try:
                return self._call(req)
            except AuthenticationError:
                if self._refresh_auth():
                    self._save_cookies_to_file()
                    return self._call(req)
                else:
                    raise AuthenticationError("Authentication failed and could not be refreshed.")

    def _call(self, req: device_pb2.Request) -> device_pb2.Response:
        """
        Internal method to handle the gRPC-Web request.

        Parameters:
            req (device_pb2.Request): The request to send.

        Returns:
            device_pb2.Response: The server response.
        """
        body_bytes = req.SerializeToString(deterministic=True)
        prefix = (len(body_bytes)).to_bytes(5, byteorder='big', signed=True)
        body = prefix + body_bytes
        headers = {
            "X-Grpc-Web": "1",
            "X-User-Agent": "okhttp/4.9.2",
            "Content-Type": "application/grpc-web+proto",
            "Accept-Encoding": "gzip, deflate, br",
            "x-xsrf-token": self._xsrf_token,
            "cookie": self._cookie,
        }
        
        response = self._client.post(self._url, content=body, headers=headers, timeout=DEFAULT_TIMEOUT)
        if response.status_code == 200:
            resp_bytes = response.content
            if len(resp_bytes) == 0:
                grpc_msg = response.headers.get("Grpc-Message", "")
                raise ResponseError(f"gRPC error msg: {grpc_msg}" if grpc_msg else "Empty response")
            content_bytes = self._parse_grpc_web_response(resp_bytes)
            resp_msg = device_pb2.Response()
            resp_msg.ParseFromString(content_bytes)
            return resp_msg
        elif response.status_code == 401:
            raise AuthenticationError("Unauthorized")
        else:
            raise ResponseError(f"HTTP error {response.status_code}: {response.reason_phrase}")

    @staticmethod
    def _parse_grpc_web_response(resp_bytes):
        """
        Parse the gRPC-Web response to retrieve the message content.

        Parameters:
            resp_bytes (bytes): Raw response bytes.

        Returns:
            bytes: Parsed message content.
        """
        idx = 0
        messages = []
        while idx < len(resp_bytes):
            flag = resp_bytes[idx]
            idx += 1
            msg_len = int.from_bytes(resp_bytes[idx:idx + 4], byteorder='big')
            idx += 4
            msg = resp_bytes[idx:idx + msg_len]
            idx += msg_len
            if flag == 0x00:
                messages.append(msg)
            elif flag == 0x80:
                break
            else:
                raise ResponseError(f"Unknown gRPC-Web frame flag: {flag}")
        return b''.join(messages)

    @contextmanager
    def _timeout(self, seconds: int = DEFAULT_TIMEOUT):
        """Context manager for setting a request timeout."""
        start = time.time()
        yield
        if time.time() - start > seconds:
            raise TimeoutError(ERR_TIMEOUT)

    def get_account(self) -> Account:
        """
        Retrieve the account object, refreshing authentication if necessary.

        Returns:
            Account: The account information.
        """
        if self._account is None:
            self._refresh_auth()
        if self._account is None:
            raise ValueError("Unable to retrieve account")
        return self._account

    def _refresh_auth(self) -> None:
        """
        Refresh the authentication tokens if expired.

        Returns:
            bool: True if authentication was successfully refreshed, False otherwise.
        """
        with self._lock:
            headers = {
                "X-User-Agent": "okhttp/4.9.2",
                "Accept": "application/json",                
                "Cookie": self._cookie,
            }
            try:
                response = self._client.get(self._auth_url, headers=headers, timeout=DEFAULT_TIMEOUT)
            except httpx.RequestError as e:
                raise AuthenticationError(f"Error during authentication refresh request: {e}")

            if response.status_code != 200:
                raise AuthenticationError(f"Authentication failed with status code {response.status_code}")

            try:
                self._account = Account.model_validate_json(response.text)
            except json.JSONDecodeError as e:
                raise AuthenticationError(f"Failed to parse account response: {e}")

            for cookie in response.cookies.keys():
                self._cookie_jar.delete(cookie)

            for cookie in response.cookies.jar:
                self._cookie_jar.set(
                    name=cookie.name,
                    value=str(cookie.value),
                    domain=cookie.domain,
                    path=cookie.path,
                )

            self._xsrf_token = self._cookie_jar.get("XSRF-TOKEN", "")

            if not self._xsrf_token:
                raise AuthenticationError("Failed to retrieve XSRF token")



import json


def parse_cookie_json(cookie_json: str) -> str:
    try:
        cookies = json.loads(cookie_json)
        cookie_string = "; ".join(
            f"{cookie['name']}={cookie['value']}" for cookie in cookies)

        return cookie_string
    except (json.JSONDecodeError, KeyError) as e:
        raise ValueError(f"Error al procesar el JSON de cookies: {e}")


from build.lib.starlink_client.wifi_config import CurrentWifiConfig
from spacex.api.device.device_pb2 import Request
from spacex.api.device.wifi_config_pb2 import WifiConfig, AuthWpa2
from spacex.api.device.wifi_pb2 import WifiSetConfigRequest, WifiGetStatusResponse
from starlink_client.wifi_config import NewWifiConfig


class ServiceSetBuilder:
    @classmethod
    def build_setup_wifi_request(cls, router_id: str, new_cfg: NewWifiConfig, router_status: WifiGetStatusResponse):
        """
        Build the Wi-Fi configuration request.

        Parameters:
            router_id (str): The router's ID.
            new_cfg (NewWifiConfig): The desired Wi-Fi configuration.
            router_status (device_pb2.WifiGetStatusResponse): The current Wi-Fi status.

        Returns:
            device_pb2.Request: The gRPC request message.
        """
        # Build the service sets        
        ss = cls.build_service_sets(new_cfg, router_status)

        # Create the Wi-Fi configuration request
        wifi_config = WifiConfig(
            apply_networks=True,
            networks=[
                WifiConfig.Network(
                    basic_service_sets=ss
                )
            ]
        )

        wifi_set_config_request = WifiSetConfigRequest(
            wifi_config=wifi_config
        )

        grpc_req = Request(
            target_id=router_id,
            wifi_set_config=wifi_set_config_request
        )

        return grpc_req

    @classmethod
    def build_service_sets(cls, cfg: NewWifiConfig, router_status: WifiGetStatusResponse):
        """
        Build the BasicServiceSet configurations for both 2.4GHz and 5GHz bands.

        Parameters:            
            cfg (NewWifiConfig): The desired Wi-Fi configuration.
            router_status (device_pb2.WifiGetStatusResponse): The current Wi-Fi status.

        Returns:
            List[device_pb2.WifiConfig.BasicServiceSet]: List of BasicServiceSet configurations.
        """
        # Retrieve the current Wi-Fi status and get current BSSIDs and SSIDs        
        current_cfg = cls.append_bssid_to_config(router_status)

        # Build service set list
        ss_list = cls.build_service_set_list(cfg, current_cfg)

        return ss_list

    @classmethod
    def append_bssid_to_config(cls, router_status: WifiGetStatusResponse) -> CurrentWifiConfig:
        """
        Retrieve the current Wi-Fi status of the router and update the configuration.

        Parameters:            
            router_status (device_pb2.WifiGetStatusResponse): The current Wi-Fi status.

        Returns:
            CurrentWifiConfig: The current Wi-Fi configuration with BSSIDs and SSIDs.
        """
        current_cfg = CurrentWifiConfig()

        # Iterate over the existing networks and their Basic Service Sets
        for network in router_status.config.networks:
            for bss in network.basic_service_sets:
                if bss.band == WifiConfig.RF_2GHZ:
                    current_cfg.bssid_24ghz = bss.bssid
                    current_cfg.current_ssid_24ghz = bss.ssid
                elif bss.band == WifiConfig.RF_5GHZ:
                    current_cfg.bssid_5ghz = bss.bssid
                    current_cfg.current_ssid_5ghz = bss.ssid

        return current_cfg

    @classmethod
    def build_service_set_list(cls, cfg: NewWifiConfig, current_cfg: CurrentWifiConfig):
        """
        Build the list of BasicServiceSet configurations.

        Parameters:
            cfg (NewWifiConfig): The desired Wi-Fi configuration.
            current_cfg (CurrentWifiConfig): The current Wi-Fi configuration.

        Returns:
            List[device_pb2.WifiConfig.BasicServiceSet]: List of BasicServiceSet configurations.
        """
        service_sets = []

        # Build 2.4GHz service set
        ss_24ghz = cls.build_24ghz_service_set(cfg, current_cfg)
        service_sets.append(ss_24ghz)

        # Build 5GHz service set
        ss_5ghz = cls.build_5ghz_service_set(cfg, current_cfg)
        service_sets.append(ss_5ghz)

        return service_sets

    @staticmethod
    def build_24ghz_service_set(cfg: NewWifiConfig, current_cfg: CurrentWifiConfig):
        """
        Build the BasicServiceSet configuration for the 2.4GHz band.

        Parameters:
            cfg (NewWifiConfig): The desired Wi-Fi configuration.
            current_cfg (CurrentWifiConfig): The current Wi-Fi configuration.

        Returns:
            device_pb2.WifiConfig.BasicServiceSet: The BasicServiceSet configuration for 2.4GHz.
        """
        # Use the current SSID if none is provided
        ssid_24ghz = cfg.ssid_24ghz or current_cfg.current_ssid_24ghz

        # If a new password is provided, reset BSSID
        bssid_24ghz = current_cfg.bssid_24ghz
        if cfg.password_24ghz:
            bssid_24ghz = ""

        auth_field = None
        if cfg.password_24ghz:
            auth_field = AuthWpa2(password=cfg.password_24ghz)

        ss_24ghz = WifiConfig.BasicServiceSet(
            bssid=bssid_24ghz,
            ssid=ssid_24ghz,
            band=WifiConfig.RF_2GHZ,
            hidden=cfg.hide_24ghz,
            auth_wpa2=auth_field
        )

        return ss_24ghz

    @staticmethod
    def build_5ghz_service_set(cfg: NewWifiConfig, current_cfg: CurrentWifiConfig):
        """
        Build the BasicServiceSet configuration for the 5GHz band.

        Parameters:
            cfg (NewWifiConfig): The desired Wi-Fi configuration.
            current_cfg (CurrentWifiConfig): The current Wi-Fi configuration.

        Returns:
            device_pb2.WifiConfig.BasicServiceSet: The BasicServiceSet configuration for 5GHz.
        """
        # Use the current SSID if none is provided
        ssid_5ghz = cfg.ssid_5ghz or current_cfg.current_ssid_5ghz

        # If a new password is provided, reset BSSID
        bssid_5ghz = current_cfg.bssid_5ghz
        if cfg.password_5ghz:
            bssid_5ghz = ""

        auth_wpa2 = None
        if cfg.password_5ghz:
            auth_wpa2 = AuthWpa2(password=cfg.password_5ghz)

        ss_5ghz = WifiConfig.BasicServiceSet(
            bssid=bssid_5ghz,
            ssid=ssid_5ghz,
            band=WifiConfig.RF_5GHZ,
            hidden=cfg.hide_5ghz,
            auth_wpa2=auth_wpa2
        )

        return ss_5ghz


from spacex.api.device.device_pb2 import Request, GetStatusRequest
from spacex.api.device.wifi_config_pb2 import WifiConfig
from spacex.api.device.wifi_pb2 import WifiGetStatusResponse
from starlink_client.grpc_web_base_client import AuthenticationError, ResponseError, GrpcWebBaseClient
from starlink_client.service_set_builder import ServiceSetBuilder
from starlink_client.wifi_config import NewWifiConfig


class GrpcWebClient(GrpcWebBaseClient):
    def get_wifi_status(self, router_id: str) -> WifiGetStatusResponse:
        """
        Get the Wi-Fi status for a specific router.

        Parameters:
            router_id (str): The router's ID.

        Returns:
            device_pb2.WifiConfig: Wi-Fi configuration of the router.
        """
        router_id = self.add_prefix_to_router_id(router_id)
        grpc_req = Request(target_id=router_id, get_status=GetStatusRequest())
        resp = self.call(grpc_req)
        return resp.wifi_get_status

    def setup_wifi(self, router_id: str, new_cfg: NewWifiConfig) -> WifiConfig:
        """
        Configure the Wi-Fi settings for a specified router.

        Parameters:
            router_id (str): The router's ID.
            new_cfg (NewWifiConfig): The desired Wi-Fi configuration.

        Returns:
            device_pb2.WifiConfig: The updated Wi-Fi configuration from the device.
        """
        min_password_len = 8
        max_password_len = 32

        if new_cfg.password_24ghz and not (min_password_len <= len(new_cfg.password_24ghz) <= max_password_len):
            raise ValueError("Invalid 2.4GHz password: passwords must be between 8 and 32 characters")
        if new_cfg.password_5ghz and not (min_password_len <= len(new_cfg.password_5ghz) <= max_password_len):
            raise ValueError("Invalid 5GHz password: passwords must be between 8 and 32 characters")

        # Prefix the router ID as required
        router_id = self.add_prefix_to_router_id(router_id)

        # Build the request
        router_status = self.get_wifi_status(router_id)
        grpc_req = ServiceSetBuilder.build_setup_wifi_request(
            router_id, new_cfg, router_status,
        )

        # Send the request to the device
        resp = self.call(grpc_req)

        # Process the response
        if resp.WhichOneof('response') != 'wifi_set_config':
            raise ValueError("Unexpected response type")

        # The updated Wi-Fi configuration from the device
        updated_wifi_config = resp.wifi_set_config.updated_wifi_config

        return updated_wifi_config

    @staticmethod
    def add_prefix_to_router_id(router_id: str) -> str:
        """
        Prefix the router ID as required by the client.

        Parameters:
            router_id (str): The original router ID.

        Returns:
            str: The prefixed router ID.
        """
        # Implement the prefix logic as required
        if router_id.startswith("Router-"):
            return router_id
        return f"Router-{router_id}"



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="starlink-client",
    version="0.1.4",
    packages=find_packages(include=["py.typed","starlink_client", "starlink_client.*"]),
    include_package_data=True,
    install_requires=[
        "grpcio",
        "grpcio-status",
        "proto-plus",
        "protobuf",
        "requests",
        "httpx",
        "pydantic",
        "httpx[http2]",
    ],
    description="A Python client for Starlink.",
    author="Hector Oliveros",
    author_email="hector.oliveros.leon@gmail.com",
    url="https://github.com/Eitol/starlink-client",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Typing :: Typed",
    ],
    python_requires='>=3.9',
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="starlink client grpc satellite internet antenna",
)



from typing import List

from pydantic import BaseModel, EmailStr


class Permission(BaseModel):
    featureAccess: int
    permission: int


class Account(BaseModel):
    email: str
    emailVerified: bool
    familyName: str
    givenName: str
    locale: str
    name: str
    subjectId: str
    accountId: str
    updatedAt: int
    isSupportAgent: bool
    isSpacexEmployee: bool
    enabled: bool
    canManageClients: bool
    roles: List[int]
    employeeAccountPermissions: List[str]
    permissions: List[Permission]
